@@ -4,13 +4,27 @@ import warnings
 
 import pkg_resources
 import pytest
-from _pytest.resultlog import ResultLog
 from _pytest.runner import runtestprotocol
 
+HAS_RESULTLOG = False
+
+try:
+    from _pytest.resultlog import ResultLog
+
+    HAS_RESULTLOG = True
+except ImportError:
+    # We have a pytest >= 6.1
+    pass
+
+
 PYTEST_GTE_54 = pkg_resources.parse_version(
     pytest.__version__
 ) >= pkg_resources.parse_version("5.4")
 
+PYTEST_GTE_63 = pkg_resources.parse_version(
+    pytest.__version__
+) >= pkg_resources.parse_version("6.3.0.dev")
+
 
 def works_with_current_xdist():
     """Returns compatibility with installed pytest-xdist version.
@@ -72,7 +86,9 @@ def pytest_configure(config):
 
 
 def _get_resultlog(config):
-    if PYTEST_GTE_54:
+    if not HAS_RESULTLOG:
+        return None
+    elif PYTEST_GTE_54:
         # hack
         from _pytest.resultlog import resultlog_key
 
@@ -82,7 +98,9 @@ def _get_resultlog(config):
 
 
 def _set_resultlog(config, resultlog):
-    if PYTEST_GTE_54:
+    if not HAS_RESULTLOG:
+        pass
+    elif PYTEST_GTE_54:
         # hack
         from _pytest.resultlog import resultlog_key
 
@@ -92,7 +110,7 @@ def _set_resultlog(config, resultlog):
 
 
 # making sure the options make sense
-# should run before / at the begining of pytest_cmdline_main
+# should run before / at the beginning of pytest_cmdline_main
 def check_options(config):
     val = config.getvalue
     if not val("collectonly"):
@@ -161,6 +179,16 @@ def get_reruns_delay(item):
     return delay
 
 
+def get_reruns_condition(item):
+    rerun_marker = _get_marker(item)
+
+    condition = True
+    if rerun_marker is not None and "condition" in rerun_marker.kwargs:
+        condition = rerun_marker.kwargs["condition"]
+
+    return condition
+
+
 def _remove_cached_results_from_failed_fixtures(item):
     """
     Note: remove all cached_result attribute from every fixture
@@ -181,15 +209,17 @@ def _remove_cached_results_from_failed_fixtures(item):
 
 def _remove_failed_setup_state_from_session(item):
     """
-    Note: remove all _prepare_exc attribute from every col in stack of
-          _setupstate and cleaning the stack itself
+    Note: remove all failures from every node in _setupstate stack
+          and clean the stack itself
     """
-    prepare_exc = "_prepare_exc"
-    setup_state = getattr(item.session, "_setupstate")
-    for col in setup_state.stack:
-        if hasattr(col, prepare_exc):
-            delattr(col, prepare_exc)
-    setup_state.stack = list()
+    setup_state = item.session._setupstate
+    if PYTEST_GTE_63:
+        setup_state.stack = {}
+    else:
+        for node in setup_state.stack:
+            if hasattr(node, "_prepare_exc"):
+                del node._prepare_exc
+        setup_state.stack = []
 
 
 def _should_hard_fail_on_error(session_config, report):
@@ -207,6 +237,19 @@ def _should_hard_fail_on_error(session_config, report):
     return True
 
 
+def _should_not_rerun(item, report, reruns):
+    xfail = hasattr(report, "wasxfail")
+    is_terminal_error = _should_hard_fail_on_error(item.session.config, report)
+    condition = get_reruns_condition(item)
+    return (
+        item.execution_count > reruns
+        or not report.failed
+        or xfail
+        or is_terminal_error
+        or not condition
+    )
+
+
 def pytest_runtest_protocol(item, nextitem):
     """
     Note: when teardown fails, two reports are generated for the case, one for
@@ -223,7 +266,7 @@ def pytest_runtest_protocol(item, nextitem):
     # first item if necessary
     check_options(item.session.config)
     delay = get_reruns_delay(item)
-    parallel = hasattr(item.config, "slaveinput")
+    parallel = hasattr(item.config, "slaveinput") or hasattr(item.config, "workerinput")
     item.execution_count = 0
 
     need_to_run = True
@@ -233,15 +276,8 @@ def pytest_runtest_protocol(item, nextitem):
         reports = runtestprotocol(item, nextitem=nextitem, log=False)
 
         for report in reports:  # 3 reports: setup, call, teardown
-            is_terminal_error = _should_hard_fail_on_error(item.session.config, report)
             report.rerun = item.execution_count - 1
-            xfail = hasattr(report, "wasxfail")
-            if (
-                item.execution_count > reruns
-                or not report.failed
-                or xfail
-                or is_terminal_error
-            ):
+            if _should_not_rerun(item, report, reruns):
                 # last run or no failure detected, log normally
                 item.ihook.pytest_runtest_logreport(report=report)
             else:
@@ -267,15 +303,13 @@ def pytest_runtest_protocol(item, nextitem):
 
 
 def pytest_report_teststatus(report):
-    """Adapted from https://pytest.org/latest/_modules/_pytest/skipping.html
-    """
+    """Adapted from https://pytest.org/latest/_modules/_pytest/skipping.html"""
     if report.outcome == "rerun":
         return "rerun", "R", ("RERUN", {"yellow": True})
 
 
 def pytest_terminal_summary(terminalreporter):
-    """Adapted from https://pytest.org/latest/_modules/_pytest/skipping.html
-    """
+    """Adapted from https://pytest.org/latest/_modules/_pytest/skipping.html"""
     tr = terminalreporter
     if not tr.reportchars:
         return
@@ -296,35 +330,37 @@ def show_rerun(terminalreporter, lines):
     if rerun:
         for rep in rerun:
             pos = rep.nodeid
-            lines.append("RERUN {}".format(pos))
+            lines.append(f"RERUN {pos}")
 
 
-class RerunResultLog(ResultLog):
-    def __init__(self, config, logfile):
-        ResultLog.__init__(self, config, logfile)
+if HAS_RESULTLOG:
 
-    def pytest_runtest_logreport(self, report):
-        """
-        Adds support for rerun report fix for issue:
-        https://github.com/pytest-dev/pytest-rerunfailures/issues/28
-        """
-        if report.when != "call" and report.passed:
-            return
-        res = self.config.hook.pytest_report_teststatus(report=report)
-        code = res[1]
-        if code == "x":
-            longrepr = str(report.longrepr)
-        elif code == "X":
-            longrepr = ""
-        elif report.passed:
-            longrepr = ""
-        elif report.failed:
-            longrepr = str(report.longrepr)
-        elif report.skipped:
-            longrepr = str(report.longrepr[2])
-        elif report.outcome == "rerun":
-            longrepr = str(report.longrepr)
-        else:
-            longrepr = str(report.longrepr)
+    class RerunResultLog(ResultLog):
+        def __init__(self, config, logfile):
+            ResultLog.__init__(self, config, logfile)
 
-        self.log_outcome(report, code, longrepr)
+        def pytest_runtest_logreport(self, report):
+            """
+            Adds support for rerun report fix for issue:
+            https://github.com/pytest-dev/pytest-rerunfailures/issues/28
+            """
+            if report.when != "call" and report.passed:
+                return
+            res = self.config.hook.pytest_report_teststatus(report=report)
+            code = res[1]
+            if code == "x":
+                longrepr = str(report.longrepr)
+            elif code == "X":
+                longrepr = ""
+            elif report.passed:
+                longrepr = ""
+            elif report.failed:
+                longrepr = str(report.longrepr)
+            elif report.skipped:
+                longrepr = str(report.longrepr[2])
+            elif report.outcome == "rerun":
+                longrepr = str(report.longrepr)
+            else:
+                longrepr = str(report.longrepr)
+
+            self.log_outcome(report, code, longrepr)
